import click
from shutil import copyfile
import re
import subprocess
import sys


def rewrite_if_bind(snakemake_other):
    """
    Function to rewrite --bind params
    It modifies click.UNPROCESSED
    """
    bind_args = list(filter(re.compile(".*--bind.*").match, snakemake_other))  # Try to select --bind
    if bind_args:
        bind_args_rewrite = f'--apptainer-args "--bind {bind_args[0].split(" ")[1]}"'
        snakemake_other_list = list(filter(lambda x: x not in [bind_args[0], "--apptainer-args"], snakemake_other))  # remove value to rewrite
        snakemake_other_list.insert(0, bind_args_rewrite)
        return snakemake_other_list
    else:
        return snakemake_other


def build_pdf(cmd_snakemake_base):
    dag_cmd_snakemake = f"{cmd_snakemake_base} --dag | dot -Tpdf > schema_pipeline_dag.pdf"
    click.secho(f"    {dag_cmd_snakemake}\n", fg='bright_blue')
    process = subprocess.run(dag_cmd_snakemake, shell=True, check=False, stdout=sys.stdout, stderr=sys.stderr)
    if int(process.returncode) >= 1:
        raise SystemExit
    rulegraph_cmd_snakemake = f"{cmd_snakemake_base} --rulegraph | dot -Tpdf > schema_pipeline_global.pdf"
    click.secho(f"    {rulegraph_cmd_snakemake}\n", fg='bright_blue')
    process = subprocess.run(rulegraph_cmd_snakemake, shell=True, check=False, stdout=sys.stdout, stderr=sys.stderr)
    if int(process.returncode) >= 1:
        raise SystemExit
    filegraph_cmd_snakemake = f"{cmd_snakemake_base} --filegraph | dot -Tpdf > schema_pipeline_files.pdf"
    click.secho(f"    {filegraph_cmd_snakemake}\n", fg='bright_blue')
    process = subprocess.run(filegraph_cmd_snakemake, shell=True, check=False, stdout=sys.stdout, stderr=sys.stderr)
    if int(process.returncode) >= 1:
        raise SystemExit


def __run(snake_installer, configfile, pdf, snakemake_other):
    """
    \b
    Run snakemake command line with mandatory parameters.
    SNAKEMAKE_OTHER: You can also pass additional Snakemake parameters
    using snakemake syntax.
    These parameters will take precedence over Snakemake ones, which were
    defined in the profile.
    See: https://snakemake.readthedocs.io/en/stable/executing/cli.html

    Example:
        rattleSNP run -c configfile.yaml --dry-run --jobs 200
        rattleSNP run -c configfile.yaml --threads 8 --dry-run
        rattleSNP run -c configfile.yaml --apptainer-args '--bind /mnt:/mnt'

    """
    if not snake_installer.user_profile_config.exists():
        click.secho(f"    Please run command line '{snake_installer.soft_name} edit_profile' before the first run of {snake_installer.soft_name} see {snake_installer.docs}", fg="red", err=True)
        exit()
    profile = snake_installer.get_active_profile
    tools = snake_installer.get_active_tools_path

    # get user arguments
    click.secho(f'    Profile Path: {profile}', fg='yellow')
    click.secho(f'    configfile file: {configfile}', fg='yellow')
    click.secho(f'    Tools Path file: {tools}', fg='yellow')

    cmd_snakemake_base = f"snakemake -s {snake_installer.snakefile} --configfile {configfile} --workflow-profile {profile} {' '.join(rewrite_if_bind(snakemake_other))}"
    click.secho(f"\n    {cmd_snakemake_base}\n", fg='bright_blue')
    process = subprocess.run(cmd_snakemake_base, shell=True, check=False, stdout=sys.stdout, stderr=sys.stderr)
    if int(process.returncode) >= 1:
        raise SystemExit(f"""ERROR when running {snake_installer.soft_name} please see logs file.
If need you can open issue on your repository {snake_installer.git_url}, 
but please provide information about your installation with command: '{snake_installer.soft_name} --install_env'""")
    if pdf:
        build_pdf(cmd_snakemake_base)



from .snake_wrapper import SnakEcdysis, SnakeInstaller
from .cli_wrapper import main_wrapper
from .useful_function import *
from ._version import version as __version__
from ._version import version_tuple

__doc__ = """
Are you looking for a simplified installation process for your Snakemake workflows, including the various Python packages and the multitude of tools used by your pipelines? 

Would you like to simplify the use of your workflows with user-friendly commands and subcommands that even non-bioinformatician users can easy use? Look no further - Snakecdysis is the solution for you!

"""


dico_tool = {
        "soft_path":  Path(__file__).resolve().parent.joinpath("templates", "PKGNAME"),
        "url": "https://forge.ird.fr/phim/sravel/snakecdysis",
        "docs": "https://snakecdysis.readthedocs.io/en/latest/index.html",
        "description_tool": """ 
    Welcome to Snakecdysis version: VERSION! Created on January 2022
    @author: Sebastien Ravel (CIRAD),
    @email: sebastien.ravel@cirad.fr
    Please cite our github: GIT_URL
    Licencied under CeCill-C (http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html)
    and MIT Intellectual property belongs to CIRAD and authors.
    Documentation avail at: DOCS""",
        "apptainer_url_files": [('http://nas-bgpi.myds.me/DOC/rattleSNP/Singularity.rattleSNP_tools.sif',
                                   'INSTALL_PATH/containers/Singularity.rattleSNP_tools.sif'),
                                  ],
        "datatest_url_files": (
            "http://nas-bgpi.myds.me/DOC/rattleSNP/data_test_rattleSNP.zip", "data_test_rattleSNP.zip"),
    "snakemake_scripts": Path(__file__).resolve().parent.joinpath("templates", "PKGNAME", "snakemake_scripts")
    }


# import rich_click as click
import click
import sys
from pathlib import Path
from .snake_wrapper import SnakeInstaller
from .useful_function import __command_required_option_from_option, __replace_package_name, __add_header
from .global_variable import __dict_context_settings

from .install import __install, __test_install
from .edit_files import __edit_tools, __edit_profile, __create_config, __show_tools
from .run import __run


def main_wrapper(soft_path=None, url=None, docs=None, description_tool=None, apptainer_url_files=None, datatest_url_files=None, **kargs) -> click.Group:
    """Use to wrapped snakemake workflow

        Args:
            - soft_path (str): The path of the wrapped workflow installation.
            - url (str): URL of versioning repository (GitHub or GitLab).
            - docs (str): URL of documentation.
            - description_tool (str): The header printed on the terminal when running the program.
                Please include string values 'VERSION', 'GIT_URL', and 'DOCS', and the wrapper automatically replaces them with the correct values.
            - apptainer_url_files (list(tuple())): List of tuples with downloaded URL and install destination with INSTALL_PATH, like INSTALL_PATH/containers/Apptainer.CulebrONT_tools.sif.
            - datatest_url_files (tuple): Tuple with 2 values, first the URL of datatest, second download name.
            - snakefile (str): Path to the main Snakemake script file (default: snakefiles/snakefile).
            - snakemake_scripts (str): Path to the main Snakemake script file (default: snakefiles/snakefile).
            - default_profile (str): Path to create the cluster 'default_profile' directory (default: default_profile/).
            - git_configfile_path (str): Path to the default configfile YAML (default: install_files/configfile.yaml).
            - git_tools_path (str): Path to the default tools config YAML (default: install_files/tools_path.yaml).
            - git_profile_config (str): Path to the default profile config file (default: install_files/config.yaml).
            - tools_version_path (str): Path to the csv with soft and command to get version (default: snakemake_scripts/report_template/versions.csv).
            - slurm_mode_choice (list): List of snakemake value for option --software-deployment-method (default: ['env-modules', 'apptainer', 'conda', "env-modules, conda"]).
            - local_mode_choice (list): List of snakemake value for option --software-deployment-method (default: ['apptainer']).

        Return:
            click.Group (``click.Group``): all sub-commands of workflow deployement and run

        Exemple:
            >>> from snakecdysis import main_wrapper
            >>> from podiumASM import dico_tool
            >>> main = main_wrapper(**dico_tool)
            >>> main()
            >>> main = main_wrapper(soft_path="path/to/install",
                                    url="http://Snakecdysis.com",
                                    docs=docs, description_tool=description_tool,
                                    apptainer_url_files=[('http://nas-bgpi.myds.me/DOC/rattleSNP/Apptainer.rattleSNP_tools.sif',
                                   'INSTALL_PATH/containers/Apptainer.rattleSNP_tools.sif'),
                                  ('http://nas-bgpi.myds.me/DOC/rattleSNP/Apptainer.report.sif',
                                   'INSTALL_PATH/containers/Apptainer.report.sif')
                                    datatest_url_files=("http://nas-bgpi.myds.me/DOC/rattleSNP/data_test_rattleSNP.zip", "data_test_rattleSNP.zip")
    """

    snake_installer = SnakeInstaller(soft_path=soft_path, url=url, docs=docs, description_tool=description_tool, apptainer_url_files=apptainer_url_files, datatest_url_files=datatest_url_files, **kargs)

    # Create click group for all subcommand line
    @click.group(name=f"{snake_installer.soft_name}", context_settings=__dict_context_settings,
                 invoke_without_command=True, no_args_is_help=False)
    @click.option('--restore', '-r', is_flag=True, required=False, default=False, show_default=True,
                  help='Restore previous installation to use again "install" command')
    @click.option('--install_env', '-e', is_flag=True, required=False, default=False, show_default=True,
                  help='print Install path, Tools config, Install mode, Tools install mode, Current version, Latest version avail, Snakecdysis version')
    @click.version_option(snake_installer.version, "-v", "--version", message="%(prog)s, version %(version)s")
    @click.pass_context
    @__add_header(snake_installer=snake_installer)
    def main_command(ctx, restore, install_env):
        """"""
        if ctx.invoked_subcommand is None and restore:  # and check_privileges():
            if snake_installer.install_mode in ["local", "cluster"]:
                snake_installer.install_mode_file.unlink(missing_ok=False)
                snake_installer.clean_home()
                click.secho(
                    f"\n    Remove installation mode, now run:\n    {snake_installer.soft_name} install_local or "
                    f"install_cluster\n\n", fg="yellow")
            else:
                click.secho(
                    f"\n    No reset need, {snake_installer.soft_name} not install !!!!!\n    Please run: "
                    f"{snake_installer.soft_name} install_local or install_cluster !!!!\n\n",
                    fg="red")
        elif ctx.invoked_subcommand is None and install_env:
            click.secho(f"""{snake_installer.soft_name} information's to help debug:
    - Install path:\t\t{snake_installer.install_path}
    - Tools config file: \t{snake_installer.get_active_tools_path}
    - Active profile:   \t{snake_installer.get_active_profile}
    - Install mode is:   \t{snake_installer.install_mode}
    - Tools install mode is: \t{snake_installer.tools_mode}
    - Current version is: \t{snake_installer.version}
    - Latest version avail is: \t{snake_installer.latest_version}
    - Snakecdysis version: \t{snake_installer.snakecdysis_version}
                  """,  fg="bright_blue")

        elif ctx.invoked_subcommand is None and (not install_env or not restore):
            click.echo(ctx.get_help())
        if not snake_installer.user_profile_config.exists() and snake_installer.install_mode == "cluster":
            click.secho(
                f"\n  Please run command line '{snake_installer.soft_name} edit_profile' before the first run of {snake_installer.soft_name} see {snake_installer.docs}\n\n",
                fg="red", err=True)

    ############################################
    # Install mode subcommand
    ############################################
    def choose_env(ctx, param, value):
        """
        Set the environment based on the mode specified in ctx.params.

        Args:
            ctx (object): The context object containing parameters.
            param (str): The parameter name.
            value (str): The value to set the environment to.

        Returns:
            str: The value of the environment set.

        Raises:
            click.BadParameter: If the specified value is not in the allowed choices.
        """
        # Define choices based on the mode
        mode = ctx.params.get('mode')
        choices = snake_installer.slurm_mode_choice if mode == 'slurm' else snake_installer.local_mode_choice if mode == 'local' else set(
            snake_installer.slurm_mode_choice + snake_installer.local_mode_choice)
        # Check if value is empty or parsing is resilient
        if not value or ctx.resilient_parsing:
            return
        # Check if the value is in the allowed choices
        if value not in choices:
            raise click.BadParameter(f"Unexpected value '{value}' with --mode '{mode}', Allowed values are: {choices}")
        # Set the environment value in ctx.params
        ctx.params['env'] = value
        return value

    # GLOBAL
    @click.command("install", short_help=f'Install {snake_installer.soft_name} on HPC cluster or on local mode',
                   context_settings=__dict_context_settings, no_args_is_help=False)
    @click.option('--mode', '-m', default="slurm", type=click.Choice(['slurm', 'local'], case_sensitive=False),
                  show_default=True, help='Mode for installation')
    @click.option('--env', '-e', default="apptainer",
                  type=click.Choice(set(snake_installer.slurm_mode_choice + snake_installer.local_mode_choice), case_sensitive=False),
                  show_default=True,
                  help=f'Mode for tools dependencies for slurm: {snake_installer.slurm_mode_choice}, local: {snake_installer.local_mode_choice}',
                  callback=choose_env)
    @click.option('--bash_completion/--no-bash_completion', is_flag=True, required=False, default=True,
                  show_default=True,
                  help=f'Allow bash completion of {snake_installer.soft_name} commands on the bashrc file')
    @click.pass_context
    def install(ctx, mode, env, bash_completion):
        """Run installation of tool for HPC cluster"""
        __install(snake_installer, mode, env, bash_completion)

    # TEST INSTALL
    @click.command("test_install",
                   short_help=f'Test {snake_installer.soft_name} { snake_installer.install_mode if snake_installer.install_mode in ["slurm","local"] else "local or slurm"} mode with "data_test"',
                   context_settings=__dict_context_settings, no_args_is_help=False)
    @click.option('--data_dir', '-d', default=None,
                  type=click.Path(exists=False, file_okay=False, dir_okay=True, readable=True, resolve_path=True),
                  required=True, show_default=False,
                  help='Path to download data test and create config.yaml to run test')
    @__replace_package_name(snake_installer=snake_installer)
    def test_install(data_dir):
        """Test_install function downloads a scaled data test, writes a configuration file adapted to it and
        proposes a command line already to run !!!"""
        __test_install(snake_installer, data_dir)

    ############################################
    # Run mode subcommand
    ############################################

    @click.command("run", short_help='Run workflow',
                   context_settings=__dict_context_settings, no_args_is_help=True)
    @click.option('--configfile', '-c', type=click.Path(exists=True, file_okay=True, readable=True, resolve_path=True),
                  required=True, show_default=True, help=f'Configuration file for run tool')
    @click.option('--pdf', '-pdf', is_flag=True, required=False, default=False, show_default=True,
                  help='Run snakemake with --dag, --rulegraph and --filegraph')
    @click.argument('snakemake_other', nargs=-1, type=click.UNPROCESSED)
    @__replace_package_name(snake_installer=snake_installer)
    def run(configfile, pdf, snakemake_other):
        """
        \b
        Run snakemake command line with mandatory parameters.
        SNAKEMAKE_OTHER: You can also pass additional Snakemake parameters using snakemake syntax.
        These parameters will take precedence over Snakemake ones, which were defined in the profile.
        See: https://snakemake.readthedocs.io/en/stable/executing/cli.html
        \b
        Example:
            rattleSNP run -c configfile.yaml --dry-run --jobs 200
            rattleSNP run -c configfile.yaml --threads 8 --dry-run
            rattleSNP run -c configfile.yaml --apptainer-args '--bind /mnt:/mnt'
        """
        __run(snake_installer, configfile, pdf, snakemake_other)

    ############################################
    # EDIT CONFIG subcommand
    ############################################

    @click.command("edit_tools", short_help='Edit own tools version', no_args_is_help=False)
    @click.option('--restore', '-r', is_flag=True, required=False, default=False, show_default=True,
                  help='Restore default tools_config.yaml (from install)')
    @__replace_package_name(snake_installer=snake_installer)
    def edit_tools(restore):
        """"""
        __edit_tools(snake_installer, restore)

    @click.command("create_config", short_help='Create configfile.yaml for run', no_args_is_help=True)
    @click.option('--configfile', '-c', default=None,
                  type=click.Path(exists=False, file_okay=True, dir_okay=False, readable=True, resolve_path=True),
                  required=True, show_default=True, help='Path to create configfile.yaml')
    @__replace_package_name(snake_installer=snake_installer)
    def create_config(configyaml):
        """"""
        __create_config(snake_installer, configyaml)

    @click.command("edit_profile", short_help='Edit config.yaml use by profile', no_args_is_help=False)
    @__replace_package_name(snake_installer=snake_installer)
    def edit_profile():
        """"""
        __edit_profile(snake_installer)

    @click.command("show_tools", short_help='show tools version', no_args_is_help=False)
    @__replace_package_name(snake_installer=snake_installer)
    def show_tools():
        """"""
        __show_tools(snake_installer)


    # Hack for build docs with unspecified install
    args = str(sys.argv)
    if "sphinx" in args:
        main_command.add_command(run)
        main_command.add_command(edit_profile)
        main_command.add_command(create_config)
        main_command.add_command(edit_tools)
        main_command.add_command(install)
        main_command.add_command(test_install)
        main_command.add_command(show_tools)
    else:
        if snake_installer.install_mode == "slurm" or snake_installer.install_mode == "local":
            main_command.add_command(edit_profile)
            if snake_installer.user_profile_config.exists():
                main_command.add_command(run)
                main_command.add_command(test_install)
                main_command.add_command(create_config)
                main_command.add_command(edit_tools)
                main_command.add_command(show_tools)
        else:
            main_command.add_command(install)
    return main_command



import pandas as pd
import yaml
import pprint
import os
import re
import subprocess
import click
from pathlib import Path
from shutil import rmtree
from collections import OrderedDict
from click import progressbar, secho
from snakemake.common.configfile import load_configfile
from .useful_function import multiprocessing_download


class SnakeInstaller:
    """
    This class is used to install Snakemake workflows.
    """

    def __init__(self, soft_path=None, url=None, docs=None, description_tool=None, apptainer_url_files=None,
                 datatest_url_files=None, **kargs):
        """
        Initialize SnakeInstaller.

        Args:
        - soft_path (str): The path of the wrapped workflow installation.
        - url (str): URL of versioning repository (GitHub or GitLab).
        - docs (str): URL of documentation.
        - description_tool (str): The header printed on the terminal when running the program.
            Please include string values 'VERSION', 'GIT_URL', and 'DOCS', and the wrapper automatically replaces them with the correct values.
        - apptainer_url_files (list(tuple())): List of tuples with downloaded URL and install destination with INSTALL_PATH, like INSTALL_PATH/containers/Apptainer.CulebrONT_tools.sif.
        - datatest_url_files (tuple): Tuple with 2 values, first the URL of datatest, second download name.
        - snakefile (str): Path to the main Snakemake script file (default: snakefiles/snakefile).
        - snakemake_scripts (str): Path to the main Snakemake script file (default: snakefiles/snakefile).
        - default_profile (str): Path to create the cluster 'default_profile' directory (default: default_profile/).
        - git_configfile_path (str): Path to the default configfile YAML (default: install_files/configfile.yaml).
        - git_tools_path (str): Path to the default tools config YAML (default: install_files/tools_path.yaml).
        - git_profile_config (str): Path to the default profile config file (default: install_files/config.yaml).
        - tools_version_path (str): Path to the csv with soft and command to get version (default: snakemake_scripts/report_template/versions.csv).
        - slurm_mode_choice (list): List of snakemake value for option --software-deployment-method (default: ['env-modules', 'apptainer', 'conda', "env-modules, conda"]).
        - local_mode_choice (list): List of snakemake value for option --software-deployment-method (default: ['apptainer']).
        - kargs: Additional keyword arguments.
        """

        self.tools_config = None
        self.install_path = soft_path
        self.home_config_path = Path(f"~/.config/{self.soft_name}-snake8/").expanduser()
        self.git_url = url
        self.docs = docs
        self.description_tool = description_tool
        self.apptainer_url_files = apptainer_url_files
        self.datatest_url_files = datatest_url_files
        self.dico_args = kargs
        self._latest_version = None
        self._lasted_file = None
        self.slurm_mode_choice = ['env-modules', 'apptainer', 'conda', "env-modules, conda"]
        self.local_mode_choice = ['apptainer']


        # autobuild attributs
        self.install_mode_file = self.install_path.joinpath(".mode.txt")
        self.bash_completion_script = Path(f"{self.install_path}/{self.soft_name}-complete.sh")

        # Snakemake settings
        self.snakefile = self.install_path.joinpath("snakefiles", "snakefile") if "snakefile" not in self.dico_args else \
            self.dico_args["snakefile"]
        self.snakemake_scripts = self.install_path.joinpath("scripts",
                                                            "snakemake_only") if "snakemake_scripts" not in self.dico_args else \
            self.dico_args["snakemake_scripts"]
        self.default_profile = self.install_path.joinpath(
            "default_profile") if "default_profile" not in self.dico_args else self.dico_args["default_profile"]
        self.git_configfile_path = self.install_path.joinpath("install_files",
                                                          "configfile.yaml") if "git_configfile_path" not in self.dico_args else \
            self.dico_args["git_configfile_path"]

        # Tools settings
        self.git_tools_path = self.install_path.joinpath("install_files",
                                                         "tools_path.yaml") if "git_tools_path" not in self.dico_args else \
            self.dico_args["git_tools_path"]

        self.tools_version_path = self.snakemake_scripts.joinpath("report_template",
                                                                  "versions.csv") if "tools_version_path" not in self.dico_args else \
            self.dico_args["tools_version_path"]

        # cluster settings
        self.git_profile_config = self.install_path.joinpath("install_files",
                                                             "config.yaml") if "git_profile_config" not in self.dico_args else \
            self.dico_args["git_profile_config"]

        self.slurm_mode_choice = self.slurm_mode_choice if "slurm_mode_choice" not in self.dico_args else \
            self.dico_args["slurm_mode_choice"]
        self.local_mode_choice = self.local_mode_choice if "local_mode_choice" not in self.dico_args else \
            self.dico_args["local_mode_choice"]

        self._lasted_file = self.user_tools_path.parent.joinpath('.latest.txt')


    @property
    def soft_name(self) -> str:
        """The wrapped workflow name."""
        return self._install_path.stem

    @property
    def install_path(self) -> Path:
        """The path of wrapped workflow installation."""
        return Path(self._install_path)

    @install_path.setter
    def install_path(self, path: Path = None):
        """
        Set the installation path.

        Args:
        - path (Path): The installation path.
        """
        if not path:
            raise ValueError("ERROR 'soft_path' is empty but mandatory.")
        elif not Path(path).exists():
            raise NotADirectoryError(f"ERROR 'soft_path', the path '{path}' doesn't exist")
        elif not Path(path).is_absolute():
            raise ValueError(f"ERROR 'soft_path', the path '{path}' is not on absolute path")
        else:
            self._install_path = Path(path)

    @property
    def git_url(self) -> str:
        """Url of versioning repository (GitHub or GitLab)"""
        return self._git_url

    @git_url.setter
    def git_url(self, url) -> str:
        if not url or not isinstance(url, str):
            raise ValueError("ERROR 'url' is empty or not a string value.")
        self._git_url = url

    @property
    def docs(self) -> str:
        """Url of documentation"""
        return self._docs

    @docs.setter
    def docs(self, docs) -> str:
        if not docs or not isinstance(docs, str):
            raise ValueError("ERROR 'docs' is empty or not a string value.")
        self._docs = docs

    @property
    def description_tool(self) -> str:
        """The header print on terminal when run programme

        Please add string values 'VERSION', 'GIT_URL' and 'DOCS' and wrapper automatically replace by the good values
        """
        return self._description_tool

    @description_tool.setter
    def description_tool(self, description_tool) -> str:
        if not description_tool or not isinstance(description_tool, str):
            raise ValueError("ERROR key 'description_tool' is empty or not a string value on 'dico_tool'.")
        self._description_tool = description_tool.replace("VERSION", self.version).replace("GIT_URL",
                                                                                           self.git_url).replace("DOCS",
                                                                                                                 self.docs)

    @property
    def apptainer_url_files(self) -> list:
        """List of tuple with downloaded url and install destination with INSTALL_PATH. like INSTALL_PATH/containers/Apptainer.CulebrONT_tools.sif"""
        return self._apptainer_url_files

    @apptainer_url_files.setter
    def apptainer_url_files(self, list_url):
        if not list_url:
            raise AttributeError("ERROR 'apptainer_url_files' is empty but mandatory")
        if not isinstance(list_url, list):
            raise ValueError(f"ERROR 'apptainer_url_files' must be a list of tuple but is {type(list_url)}")
        self._apptainer_url_files = [(url, path_install.replace("INSTALL_PATH", self.install_path.as_posix())) for
                                       url, path_install in list_url]

    @property
    def datatest_url_files(self) -> tuple:
        """Tuple with 2 values, first the url of datatest, second download name."""
        return self._datatest_url_files

    @datatest_url_files.setter
    def datatest_url_files(self, tuple_test):
        if not isinstance(tuple_test, tuple) or not len(tuple_test) == 2:
            raise ValueError("ERROR 'datatest_url_files' must be a tuple of 2 values")
        self._datatest_url_files = tuple_test

    @property
    def install_mode(self):
        """Detect install mode of the soft, can be 'No install', 'local' or 'cluster'"""
        if not self.install_mode_file.exists():
            self._install_mode = "No install"
        else:
            self._install_mode = self.install_mode_file.open("r").readline().strip()
        return self._install_mode

    @property
    def tools_mode(self):
        """Detect install mode of the tools soft, can be 'env-modules', 'apptainer', 'conda' or 'env-modules, conda'"""
        if not self.install_mode_file.exists():
            self._tools_mode = "No install"
        else:
            try:
                self._tools_mode = self.install_mode_file.open("r").readlines()[1].strip()
            except IndexError:
                self._tools_mode = f"""Not avail, please update Snakecdysis with  'python3 -m pip install snakecdysis>=1.0.0'
                        \t\tand then reinstall {self.soft_name} with
                        \t\t'{self.soft_name} -r && {self.soft_name} install'"""
        return self._tools_mode

    @property
    def version(self):
        """The current workflow version, read on VERSION file"""
        from importlib.metadata import version
        return version(self.soft_name)

    @property
    def snakecdysis_version(self):
        """The current workflow version, read on VERSION file"""
        from snakecdysis import __version__
        return __version__

    @property
    def latest_version(self):
        """The latest workflow version, read on repository"""
        self.__update_latest_file()
        return self._latest_version

    # SNAKEMAKE SETTINGS
    @property
    def snakefile(self) -> Path:
        """Path to the main snakemake file.
        Search on INSTALL_PATH/snakefiles/snakefile
        """
        return self._snakefile

    @snakefile.setter
    def snakefile(self, snake_path):
        if not Path(snake_path).exists():
            raise FileNotFoundError(f"ERROR  the snakefile '{snake_path}' doesn't exist")
        self._snakefile = Path(snake_path)

    @property
    def snakemake_scripts(self) -> Path:
        """Path to the scripts used on snakemake rules.
        Search on INSTALL_PATH/snakemake_scripts
        """
        return self._snakemake_scripts

    @snakemake_scripts.setter
    def snakemake_scripts(self, script_path) -> Path:
        if not Path(script_path).exists():
            raise FileNotFoundError(f"ERROR  the snakemake_scripts '{script_path}' doesn't exist")
        self._snakemake_scripts = Path(script_path)

    @property
    def default_profile(self) -> Path:
        """Path to the directory of config.yaml default profile.
        install on INSTALL_PATH/default_profile
        """
        return self._default_profile

    @default_profile.setter
    def default_profile(self, default_profile) -> Path:
        # if not Path(default_profile).exists():
        #     raise NotADirectoryError(f"ERROR  the default_profile '{default_profile}' doesn't exist")
        self._default_profile = Path(default_profile)

    @property
    def get_active_profile(self):
        """return the Path to current profile config"""
        if self.user_profile_config.exists():
            return self.user_profile_config.parent
        elif self.default_profile.exists():
            return self.default_profile
        else:
            return "No install"

    @property
    def git_configfile_path(self) -> Path:
        """Path to the directory of default configfile.yaml file.
        default to INSTALL_PATH/install_files/configfile.yaml
        """
        return self._git_configfile_path

    @git_configfile_path.setter
    def git_configfile_path(self, git_configfile_path) -> Path:
        if not Path(git_configfile_path).exists():
            raise FileNotFoundError(f"ERROR  the git_configfile_path '{git_configfile_path}' doesn't exist")
        self._git_configfile_path = Path(git_configfile_path)

    @property
    def git_tools_path(self) -> Path:
        """Path to the directory of default tools_path.yaml file.
        default to INSTALL_PATH/install_files/tools_path.yaml
        """
        return self._git_tools_path

    @git_tools_path.setter
    def git_tools_path(self, git_tools_path) -> Path:
        if not Path(git_tools_path).exists():
            raise FileNotFoundError(f"ERROR  the git_tools_path '{git_tools_path}' doesn't exist")
        self._git_tools_path = Path(git_tools_path)

    @property
    def user_tools_path(self) -> Path:
        """Path to the user tools path setting.
        default to ~/.config/SOFTNAME/tools_path.yaml
        """
        return self.home_config_path.joinpath("tools_path.yaml")

    @property
    def user_profile_config(self) -> Path:
        """Path to the user profile config path setting.
        default to ~/.config/SOFTNAME/config.yaml
        """
        return self.home_config_path.joinpath("config.yaml")

    def get_last_version(self) -> str:
        """Function for know the last version of program (can be GitHub or GitLab repository)
        """
        from urllib.request import urlopen
        from re import search
        try:
            if "github" in self.git_url:
                HTML = urlopen(f"{self.git_url}/tags").read().decode('utf-8')
                str_search = f"{self.git_url.replace('https://github.com', '')}/releases/tag/.*"
                last_release = search(str_search, HTML).group(0).split('"')[0].split("/")[-1]
                self._latest_version = last_release
            else:
                from gitlab import Gitlab
                project_name_with_namespace = "/".join(self.git_url.split("//")[1].split("/")[1:])
                # print(project_name_with_namespace)
                gl = Gitlab(self.git_url.replace(f'/{project_name_with_namespace}', ""))
                project = gl.projects.get(f"{project_name_with_namespace}", lazy=True)
                releases_list = project.releases.list()
                if not releases_list:
                    last_release = "There aren’t any releases"
                else:
                    last_release = releases_list[0].name
                self._latest_version = last_release
        except Exception as e:
            print(e)
            if "403" in f"{e}":
                secho(f"\n    Forbidden access to release version, please check setting on repository\n\n",
                                              fg="red")
                last_release = "Forbidden access"
                self._latest_version = last_release
            elif "404" in f"{e}":
                secho(f"\n    Project Not Found! please check setting URL:{self.git_url}\n\n",
                                              fg="red")
                last_release = "Project Not Found"
                self._latest_version = last_release
            else:
                secho(f"\n    Unable to check release:{e}\n\n",
                                              fg="red")
                last_release = "Unable"
                self._latest_version = last_release

    def __update_latest_file(self, delta_days=30):
        from datetime import datetime, timedelta
        if not Path(self._lasted_file).exists():
            self.get_last_version()
            self._lasted_file.parent.mkdir(exist_ok=True,parents=True)
            self._lasted_file.open("w").write(self._latest_version)
        else:
            # get modification time
            file_mod_time = datetime.fromtimestamp(self._lasted_file.stat().st_mtime)  # This is a datetime.datetime object!
            now = datetime.today()
            max_delay = timedelta(days=delta_days)
            if now - file_mod_time > max_delay:
                #print(f"CRITICAL:last modified on {file_mod_time}. Threshold set to {max_delay} minutes.")
                self.get_last_version()
                self._lasted_file.open("w").write(self._latest_version)
            elif self._lasted_file.exists():
                self._latest_version = self._lasted_file.open("r").read().strip()

    def get_epilog(self) -> str:
        """Function for know the last version of program (can be GitHub or GitLab repository)
           check every 30 days to skip request of if no internet connection

        Return: epilogTool print at the end of header

        """
        self.__update_latest_file()
        epilogTools = "\n"
        if str(self.version) != self._latest_version:
            if self._latest_version < str(self.version):
                epilogTools = click.style(
                    f"\n    ** NOTE: This {self.soft_name} version ({self.version}) is higher than the production version ({self._latest_version}), you are using a dev version\n",
                    fg="yellow", bold=True)
            elif self._latest_version > str(self.version) and self._latest_version != "There aren’t any releases":
                epilogTools = click.style(
                    f"\n    ** NOTE: The Latest version of {self.soft_name} {self._latest_version} is available at {self.git_url}\n",
                    fg="yellow", underline=True)
            elif self._latest_version == "There aren’t any releases":
                epilogTools = click.style(f"\n    ** NOTE: There aren’t any releases at the moment\n", fg="red",
                                          underline=False)
            else:
                epilogTools = click.style(f"\n    ** NOTE: Can't check if new release are available\n", fg="red",
                                          underline=False)
        return epilogTools

    def create_bash_completion(self):
        """Add bash completion for version > 4.4"""
        major, minor = None, None
        bashrc_file = Path("~/.bashrc").expanduser().as_posix()
        # test shell version for fix correction
        output = subprocess.run(
            ["bash", "-c", "echo ${BASH_VERSION}"], stdout=subprocess.PIPE
        )
        match = re.search(r"^(\d+)\.(\d+)\.\d+", output.stdout.decode())
        if match is not None:
            major, minor = match.groups()
        # if major < "4" or (major == "4" and minor < "4"):
        #     click.secho(f"\n    WARNING Shell completion is not supported for Bash versions older than 4.4.\n\n",
        #                 fg="red", nl=False)
        # Test if completion script exist
        if not self.bash_completion_script.exists():
            build_completion = f"_{self.soft_name.upper()}_COMPLETE=bash_source {self.soft_name} > {self.bash_completion_script.as_posix()}"
            process = subprocess.run(build_completion, shell=True, check=False, stdout=False, stderr=False)
            # fix header print
            with self.bash_completion_script.open("r") as bash_soft_open:
                new_file = ""
                start = False
                for line in bash_soft_open:
                    if f"_{self.soft_name}_completion" in line:
                        start = True
                    if start:
                        if "complete -o nosort" in line and major < "4" or (major == "4" and minor < "4"):
                            line = line.replace("-o nosort", "")
                        new_file += line
            with self.bash_completion_script.open("w") as bash_soft_open:
                bash_soft_open.write(new_file)
            # add on bashrc
            path_bash = None
            with open(bashrc_file, "r") as bash_file_read:
                for line in bash_file_read:
                    if f"{self.soft_name.upper()}" in line:
                        path_bash = bash_file_read.readline().strip()
            if path_bash:
                load = f"{path_bash[2:]}"
                if f"{self.bash_completion_script}" != load:
                    click.secho(
                        f"\n    WARNING autocompletion for {self.soft_name.upper()} already found on {bashrc_file}, with other path please fix the good:",
                        fg="red", nl=False)
                    click.secho(
                        f"\n    Load on bashrc: {load}\n    New install:    {self.bash_completion_script.as_posix()}",
                        fg='bright_red')
            else:
                with open(bashrc_file, "a") as bash_file_open:
                    append_bashrc = f"\n#Add autocompletion for {self.soft_name.upper()}\n. {self.bash_completion_script.as_posix()}"
                    bash_file_open.write(append_bashrc)
                    click.secho(f"\n    INSTALL autocompletion for {self.soft_name.upper()} on {bashrc_file} with command {append_bashrc}", fg="yellow")
                    click.secho(f"\n    TODO:  enable completion with command:\n    source ~/.bashrc", fg="cyan")

    def clean_home(self):
        """Reset home's parameters of previous installation"""
        if self.home_config_path.exists():
            rmtree(self.home_config_path.as_posix())
        if self.default_profile.exists():
            rmtree(self.default_profile.as_posix())
        if self.bash_completion_script.exists():
            self.bash_completion_script.unlink(missing_ok=True)

    def fail(self):
        """If installation fail, reset already install files"""
        rmtree(self.default_profile, ignore_errors=True)
        self.install_mode_file.unlink(missing_ok=True)
        self.clean_home()
        click.secho(f"\n    INSTALL FAIL remove already install files: {self.default_profile} ", fg="red", err=True)
        raise SystemExit

    def write_user_tools_path(self):
        """Check if file is created and then write with modification of apptainer path"""
        if not self.user_tools_path.exists():
            self.user_tools_path.parent.mkdir(parents=True, exist_ok=True)
            self.user_tools_path.write_text(
                self.git_tools_path.read_text().replace("INSTALL_PATH", self.install_path.as_posix()))

    def check_and_download_apptainer(self):
        """Download apptainer/apptainer files if provided"""
        # check if already download if true pop from list
        self.write_user_tools_path()
        SINGULARITY_URL_FILES_DOWNLOAD = []
        for url, path_install in self.apptainer_url_files:
            if not Path(path_install).exists():
                Path(path_install).parent.mkdir(parents=True, exist_ok=True)
                SINGULARITY_URL_FILES_DOWNLOAD.append((url, path_install))
                click.secho(f"    File: {path_install} is being downloaded.", fg="yellow", nl=True)
            else:
                click.secho(f"    File: {path_install} already downloaded, done.", fg="yellow", nl=True)
        multiprocessing_download(SINGULARITY_URL_FILES_DOWNLOAD)
        click.secho(
            f"\n    WARNING please check if binding is active on your apptainer/apptainer configuration, see https://apptainer.org/docs/user/main/bind_paths_and_mounts.html\n"
            f"    Or use: --apptainer-args '--bind $HOME'  on the command line",
            fg="bright_red")

    @property
    def get_active_tools_path(self):
        """return the active path of tools config"""
        if self.user_tools_path.exists():
            return self.user_tools_path
        else:
            return self.git_tools_path

    def get_tool_configfile(self):
        """Test path of tools_path.yaml on default install path, home or argument"""
        self.tools_config = load_configfile(self.get_active_tools_path)

    def tools_version_to_df(self, csv_file=None, active_tools_list=None, output_file=None):
        """check how tools install and get version to save on file"""
        if not csv_file:
            click.secho("\n    CODE error, please provide file or active_tools_list", fg="red", err=True)
            raise SystemExit
        self.get_tool_configfile()
        run_all = False
        if not active_tools_list:
            run_all = True
        df = pd.read_csv(csv_file)
        df.fillna('', inplace=True)
        df.sort_values(by=df.columns[0], axis=0, inplace=True)
        df.reset_index(inplace=True,drop=True)
        dico_version_cmd = df.to_dict(orient="index")
        dico_version_value = df.drop(columns=[df.columns[2]]).to_dict(orient="index")

        try:
            with progressbar(dico_version_cmd.items()) as bar:
                for index, dico in bar:
                    tool = dico[df.columns[0]]
                    secondary = dico[df.columns[1]]
                    cmd_version = dico[df.columns[2]]
                    #print(f"\n{tool}, {secondary} {type(secondary)}, {cmd_version}\n")
                    if run_all:
                        active_tools_list.append(tool)
                        active_tools_list.append(secondary)
                    if tool in active_tools_list or secondary in active_tools_list:
                        if self.tools_mode == "apptainer":
                            sif_path = self.tools_config["APPTAINER"]["TOOLS"]
                            run_version_cmd = f"apptainer exec {sif_path} {cmd_version}"
                        elif self.tools_mode == "env-modules":
                            tools_module = self.tools_config["ENV-MODULES"]
                            if secondary and secondary.upper() in tools_module.keys():
                                run_version_cmd = f"module purge && module load {tools_module[secondary.upper()]} && {cmd_version}"
                            else:
                                run_version_cmd = f"module purge && module load {tools_module[tool.upper()]} && {cmd_version}"

                        process = subprocess.run(run_version_cmd, shell=True, check=False, capture_output=True,
                                                 encoding="UTF-8")
                        if int(process.returncode) >= 1:
                            raise SystemExit(f"Error with command line:\n'{run_version_cmd}'\n'{process.stderr.rstrip()=}'\n{process.returncode=}\n{process.stdout.rstrip()=}")
                        else:
                            dico_version_value[index]["Version"] = process.stdout.strip()
                    else:
                        dico_version_value.pop(index)
        except UnboundLocalError as e:
            raise SystemExit(f"Error getting mode tools:\n'{self.tools_mode=}\n{self.install_path=}\n{self.install_mode_file=}\n{self.install_mode=}\n")
        except KeyError as e:
            raise SystemExit(f"Error getting Tool '{e}\n, not on avail list: {tools_module.keys()}\n")

        df_version = pd.DataFrame.from_dict(dico_version_value, orient="index")
        desired_order = [df.columns[0], df.columns[1], 'Version', df.columns[3]]
        df_version = df_version[desired_order]
        if output_file:
            # check parent dir
            Path(output_file).parent.mkdir(exist_ok=True, parents=True)
            with open(output_file, "w") as out:
                df_version.to_csv(out, sep=",", index=False)
        return df_version

    def __repr__(self):
        return f"{self.__dict__}"


class SnakEcdysis(SnakeInstaller):
    """ test generic wrapper class

        Args:
            soft_path (str): The path of wrapped workflow installation
            url (str): Url of versioning repository (GitHub or GitLab)
            docs (str): Url of documentation
            description_tool (str):The header print on terminal when run programme. Please add string values 'VERSION', 'GIT_URL' and 'DOCS' and wrapper automatically replace by th good values
            apptainer_ur_files (list(tuple()): List of tuple with downloaded url and install destination with INSTALL_PATH. like INSTALL_PATH/containers/Apptainer.CulebrONT_tools.sif
            datatest_url_files (tuple): Tuple with 2 values, first the url of datatest, second download name.

    """

    def __init__(self, dico_tool=None, workflow=None, config=None, **kargs):
        if dico_tool:
            super().__init__(**dico_tool)
        else:
            super().__init__(**kargs)
        # workflow is available only in __init__
        # print("\n".join(list(workflow.__dict__.keys())))
        # print(workflow.__dict__)
        self.tools_config = None
        if workflow.overwrite_configfiles:
            self.path_config = workflow.overwrite_configfiles[0]
        else:
            self.path_config = None
        self.config = config
        self.get_tool_configfile()

    def get_config_value(self, level1, level2=None, level3=None):
        """get value on config_file"""
        if level3:
            return self.config[level1][level2][level3]
        elif level2:
            return self.config[level1][level2]
        else:
            return self.config[level1]

    def set_config_value(self, level1, level2=None, value=None, level3=None):
        """Set config value on config_file"""
        if level3:
            self.config[level1][level2][level3] = value
        elif level2:
            self.config[level1][level2] = value
        else:
            self.config[level1] = value

    def write_config(self, path):
        """Write the corrected config file to path"""
        p = Path(path).parent
        p.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as config_open:
            config_open.write(self.export_use_yaml)

    def check_dir_or_string(self, level1, level2, mandatory=(), level3=None, check_string=False):
        """Check if the specified directory or string exists and is valid in the configuration.

            Parameters:
            - level1 (str): First level in the configuration hierarchy.
            - level2 (str): Second level in the configuration hierarchy.
            - mandatory (tuple): Tuple of mandatory items for the tool.
            - level3 (str): Optional third level in the configuration hierarchy.
            - check_string (bool): Flag to indicate whether to check for a string value.

            Raises:
            - NotADirectoryError: If the specified path does not exist or is not a valid directory.
            - ValueError: If the path is empty when it is expected to be a string.
            """
        path_value = self.get_config_value(level1=level1, level2=level2, level3=level3)
        if path_value:
            path = Path(path_value).resolve().as_posix() + "/"
            # it is a path
            if path_value != "" and "/" in path_value:
                if (not Path(path).exists() or not Path(path).is_dir()) and level2 not in ["OUTPUT"]:
                    raise NotADirectoryError(
                        f'CONFIG FILE CHECKING FAIL : in section:{level1}, {f"subsection:{level2} directory:{level3}" if level3 else f"directory:{level2}"}, "{path}" {"does not exist" if not Path(path).exists() else "is not a valid directory"}')
                else:
                    self.set_config_value(level1=level1, level2=level2, level3=level3, value=path)
            # it is not a path
            elif path_value and "/" not in path_value and check_string:
                self.set_config_value(level1=level1, level2=level2, level3=level3, value=path_value)
            # it is empty
            elif path_value and "/" not in path_value and check_string:
                raise ValueError(
                    f'CONFIG FILE CHECKING FAIL : in section:{level1}, {f"subsection:{level2} value:{level3}" if level3 else f"value:{level2}"}, "{path_value}" is empty')
        elif len(mandatory) > 0:
            raise NotADirectoryError(
                f'CONFIG FILE CHECKING FAIL : in section:{level1}, {f"subsection:{level2} directory:{level3}" if level3 else f"directory:{level2}"}, "{path_value}" {"does not exist" if not Path(path_value).exists() else "is not a valid directory"} but is mandatory for tool: "{",".join(mandatory)}"')

    def check_file_or_string(self, level1, level2, mandatory=(), level3=None, check_string=False):
        """Check if path is a file if not empty
        :return absolute path file"""
        path_value = self.get_config_value(level1=level1, level2=level2, level3=level3)
        path = Path(path_value).resolve().as_posix()
        # it is a path
        if path_value != "" and "/" in path_value:
            if not Path(path).exists() or not Path(path).is_file():
                raise FileNotFoundError(
                    f'CONFIG FILE CHECKING FAIL : in section:{level1}, {f"subsection:{level2}, file {level3}" if level3 else f"file {level2}"}, "{path}" {"does not exist" if not Path(path).exists() else "is not a valid file"}')
            else:
                self.set_config_value(level1=level1, level2=level2, level3=level3, value=path)
        # it is not a path
        elif path_value != "" and "/" not in path_value and check_string:
            self.set_config_value(level1=level1, level2=level2, level3=level3, value=path_value)
        # it is empty
        elif path_value == "" and "/" not in path_value and check_string:
            raise ValueError(
                f'CONFIG FILE CHECKING FAIL : in section:{level1}, {f"subsection:{level2}, value {level3}" if level3 else f"value:{level2}"}, "{path_value}" is empty')
        elif len(mandatory) > 0:
            raise FileNotFoundError(
                f'CONFIG FILE CHECKING FAIL : in  section:{level1} , {f"subsection:{level2}, value {level3}" if level3 else f"value:{level2}"}, "{path_value}" {"does not exist" if not Path(path_value).exists() else "is not a valid file"} but is mandatory for tool: "{",".join(mandatory)}"')

    @property
    def export_use_yaml(self):
        """Use to print a dump config.yaml with corrected parameters"""

        def represent_dictionary_order(yamldef, dict_data):
            return yamldef.represent_mapping('tag:yaml.org,2002:map', dict_data.items())

        def setup_yaml():
            yaml.add_representer(OrderedDict, represent_dictionary_order)

        setup_yaml()
        return yaml.dump(self.config, default_flow_style=False, sort_keys=False, indent=4)

    @property
    def string_to_dag(self):
        """ return command line for rule graph """
        return f"""snakemake -s {self.snakefile} --profile {self.get_active_profile} --rulegraph"""

    def __repr__(self):
        return f"{self.__class__}({pprint.pprint(self.__dict__)})"



import click
from shutil import copyfile
from pathlib import Path
import sys, os

def __edit_tools(snake_installer, restore):
    """Edit or restore the user tools_path.yaml file.

    Args:
        snake_installer: The SnakeInstaller instance.
        restore: A boolean indicating whether to restore the file.

    """
    if restore:
        if snake_installer.user_tools_path.exists():
            snake_installer.user_tools_path.unlink()
            click.secho(f"""\n    Success reset your own tools_path.yaml on path '{snake_installer.user_tools_path}'
                                  with default tools_path.yaml '{snake_installer.git_tools_path}""", fg="yellow")
    if not snake_installer.user_tools_path.exists():
        snake_installer.write_user_tools_path()
    if sys.stdin.isatty() and os.isatty(sys.stdout.fileno()):
        click.edit(require_save=True, extension='.yaml', filename=snake_installer.user_tools_path)
    click.secho(f"    Success install your own tools_path.yaml on path '{snake_installer.user_tools_path}'", fg="yellow")
    click.secho(f"    {snake_installer.soft_name} used '{snake_installer.user_tools_path}' at default now !!\n\n", fg="bright_red")


def __show_tools(snake_installer):
    """
    A function to show tools information using the snake_installer object.
    """
    click.secho(f"""\
    - Tools version was read using file: \t{snake_installer.get_active_tools_path}
    - Install mode is:   \t{snake_installer.install_mode}
    - Tools install mode is: \t{snake_installer.tools_mode}
    , please wait""", fg="yellow")
    df = snake_installer.tools_version_to_df(csv_file=snake_installer.tools_version_path, active_tools_list=[], output_file=None)
    click.secho(f"    Tools version are:\n{df.to_string(index=False)}", fg="yellow")


def __create_config(snake_installer, configyaml):
    """
    Create a config file using the provided snake_installer and configyaml.
    The function copies the git config files path to the configyaml path and allows editing of the file using click.
    It also prints success message and the command to add the config file.
    """
    configyaml = Path(configyaml)
    configyaml.parent.mkdir(parents=True, exist_ok=True)
    copyfile(snake_installer.git_configfiles_path.as_posix(), configyaml.as_posix())
    if sys.stdin.isatty() and os.isatty(sys.stdout.fileno()):
        click.edit(require_save=True, extension='.yaml', filename=configyaml.as_posix())
    click.secho(f"    Success create config file on path '{configyaml}'\n    add to command:", fg="yellow")
    mode = snake_installer.install_mode
    click.secho(f"    {snake_installer.soft_name} {'run_cluster' if mode == 'cluster' else 'run_local'} --config {configyaml}\n\n", fg="bright_blue")


def __edit_profile(snake_installer):
    f"""Edit file on {snake_installer.user_profile_config}"""
    snake_installer.user_profile_config.parent.mkdir(parents=True, exist_ok=True)
    if not snake_installer.user_profile_config.exists():
        copyfile(snake_installer.default_profile.joinpath("config.yaml").as_posix(), snake_installer.user_profile_config.as_posix())
    if sys.stdin.isatty() and os.isatty(sys.stdout.fileno()):
        click.edit(require_save=True, extension='.yaml', filename=snake_installer.user_profile_config.as_posix())
    click.secho(f"    See '{snake_installer.docs}' to adapt on your environment\n", fg="yellow")
    click.secho(f"    Success edit profile file on path '{snake_installer.user_profile_config}'\n", fg="yellow")



import click
from shutil import rmtree, unpack_archive
from .useful_function import multiprocessing_download
from pathlib import Path
import subprocess


def __install(snake_installer, mode, env, bash_completion):
    """
    Run installation of tool for HPC cluster.

    Args:
        snake_installer: The snake installer object.
        mode (str): The mode of installation (e.g. 'slurm' or 'local').
        env (str): The environment to install (e.g. 'apptainer' or 'env-modules').
        bash_completion (bool): Whether to enable bash completion.

    Raises:
        SystemExit: if Slurm is not found on the system.
    """
    # Remove previous installation
    snake_installer.clean_home()
    # Build default profile path
    default_profile = snake_installer.get_active_profile
    # Read default profile configuration
    profile_config_txt = snake_installer.git_profile_config.open("r").read()

    try:
        # Test if install has already been run
        if isinstance(default_profile, Path) and default_profile.exists() and click.confirm(
                click.style(f'    Profile "{default_profile}" exists. Do you want to remove and continue?\n\n', fg="red"),
                default=False, abort=True):
            rmtree(default_profile, ignore_errors=True)
        default_profile = snake_installer.default_profile
        default_profile.mkdir(exist_ok=True)

        # test if mode is slurm
        if mode == 'slurm':
            print(f"install on mode {mode}")
            profile_config_txt = profile_config_txt.replace("EXECUTOR", "slurm")
            # Get default partition from the system
            command = r"""sinfo -s | grep "*" | cut -d"*" -f1 """  # used admin default partition
            default_partition = subprocess.check_output(command, shell=True).decode("utf8").strip()
            if not default_partition:  # used in case of non default partition define by admin system
                command = r"""sinfo -s | cut -d" " -f1 | sed '/PARTITION/d' | head -n 1"""
                default_partition = subprocess.check_output(command, shell=True).decode("utf8").strip()
            if not default_partition:
                click.secho("    Error: Slurm was not found on your system !!", fg="red", err=True)
                snake_installer.fail()
                raise SystemExit
            profile_config_txt = profile_config_txt.replace("PARTITION", default_partition)



        # if local mode
        elif mode == 'local':
            profile_config_txt = profile_config_txt.replace("EXECUTOR", "local")
        # if apptainer activation
        if env == 'apptainer':
            profile_config_txt = profile_config_txt.replace("DEPLOYMENT", "apptainer")
            # check if already download
            snake_installer.check_and_download_apptainer()
            git_tools_file = snake_installer.git_tools_path.open("r").read()
            snake_installer.git_tools_path.open("w").write(git_tools_file.replace("INSTALL_PATH", f"{snake_installer.install_path}"))
        else:
            profile_config_txt = profile_config_txt.replace("DEPLOYMENT", "env-modules")
        # export to add bash completion
        if bash_completion:
            snake_installer.create_bash_completion()
        default_profile.joinpath("config.yaml").open("w").write(profile_config_txt)
        click.secho(f"\n    Profile is successfully installed on {default_profile}", fg="yellow")
        click.secho(
                f"    TODO: Please run command line '{snake_installer.soft_name} edit_profile' before the first run of {snake_installer.soft_name} see {snake_installer.docs}",
                fg="cyan", err=False)

        click.secho(f"\n    Congratulations, you have successfully installed {snake_installer.soft_name} !!!\n\n", fg="green", bold=True)
        snake_installer.install_mode_file.open("w").write(mode)
        snake_installer.install_mode_file.open("a").write(f"\n{env}")
    except Exception as e:
        click.secho(f"\n    ERROR : an error was detected, please check {e}", fg="red")
        snake_installer.fail()


def __install_cluster(snake_installer, env, bash_completion):
    """
     Run installation of tool for HPC cluster.
     Args:
         snake_installer: The snake installer object.
         env (str): The environment to install.
         bash_completion (bool): Whether to enable bash completion.
     Raises:
         SystemExit: if Slurm is not found on the system.
     """
    # Remove previous installation (ie @julie cluster then local)
    snake_installer.clean_home()
    # Build default profile path
    default_profile = snake_installer.default_profile

    try:
        # Test if install has already been run
        if default_profile.exists() and click.confirm(
                click.style(f'    Profile "{default_profile}" exists. Do you want to remove and continue?\n\n', fg="red"),
                default=False, abort=True):
            rmtree(default_profile, ignore_errors=True)
        default_profile.mkdir(exist_ok=True)

        # Read default cluster configuration
        default_cluster = snake_installer.git_profile_config.open("r").read()
        # Get default partition from the system
        command = r"""sinfo -s | grep "*" | cut -d"*" -f1 """   # used admin default partition
        default_partition = subprocess.check_output(command, shell=True).decode("utf8").strip()
        if not default_partition:      # used in case of non default partition define by admin system
            command = r"""sinfo -s | cut -d" " -f1 | sed '/PARTITION/d' | head -n 1"""
            default_partition = subprocess.check_output(command, shell=True).decode("utf8").strip()
        if not default_partition:
            click.secho("    Error: Slurm was not found on your system !!", fg="red", err=True)
            snake_installer.fail()
            raise SystemExit
        default_cluster = default_cluster.replace("PARTITION", default_partition)

        # if apptainer activation
        if env == 'apptainer':
            default_cluster = default_cluster.replace("DEPLOYMENT", "apptainer")
            # check if already download
            snake_installer.check_and_download_apptainer()
            git_tools_file = snake_installer.git_tools_path.open("r").read()
            snake_installer.git_tools_path.open("w").write(git_tools_file.replace("INSTALL_PATH", f"{snake_installer.install_path}"))
        else:
            default_cluster = default_cluster.replace("DEPLOYMENT", "env-modules")
        # export to add bash completion
        if bash_completion:
            snake_installer.create_bash_completion()
        default_profile.joinpath("config.yaml").open("w").write(default_cluster)
        click.secho(f"\n    Profile is success install on {default_profile}", fg="yellow")
        click.secho(
                f"    TODO: Please run command line '{snake_installer.soft_name} edit_profile' before the first run of {snake_installer.soft_name} see {snake_installer.docs}",
                fg="cyan", err=False)

        click.secho(f"\n    Congratulations, you have successfully installed {snake_installer.soft_name} !!!\n\n", fg="green", bold=True)
        snake_installer.install_mode_file.open("w").write("cluster")
        snake_installer.install_mode_file.open("a").write(f"\n{env}")
    except Exception as e:
        click.secho(f"\n    ERROR : an error was detected, please check {e}", fg="red")
        snake_installer.fail()


def __install_local(snake_installer, bash_completion):
    # rm previous install (ie @julie cluster then local)
    snake_installer.clean_home()
    # add path to download
    snake_installer.install_path.joinpath("containers").mkdir(exist_ok=True, parents=True)
    try:
        snake_installer.check_and_download_apptainer()
        # export to add bash completion
        if bash_completion:
            snake_installer.create_bash_completion()
        # good install
        click.secho(f"\n    Congratulations, you have successfully installed {snake_installer.soft_name} !!!\n\n", fg="green", bold=True)
        snake_installer.install_mode_file.open("w").write("local")
        snake_installer.install_mode_file.open("a").write("\napptainer")
    except Exception as e:
        snake_installer.install_mode_file.unlink(missing_ok=True)
        click.secho(f"\n    ERROR : an error was detected, please check {e}", fg="red")
        raise SystemExit


def __test_install(snake_installer, data_dir):
    """Test_install function downloads a scaled data test, writes a configuration file adapted to it and proposes a command line already to run !!!"""
    # create dir test and configure config.yaml
    data_dir = Path(data_dir).resolve()
    click.secho(f"\n    Created data test dir {data_dir}\n", fg="yellow")
    data_dir.mkdir(parents=True, exist_ok=True)

    data_config_path = data_dir.joinpath("data_test_config.yaml")
    click.secho(f"    Created config file to run data test: {data_config_path}\n", fg="yellow")
    txt = snake_installer.git_configfile_path.open("r").read().replace("DATA_DIR", f"{data_dir}")
    data_config_path.open("w").write(txt)

    # download data
    download_zip = data_dir.joinpath(snake_installer.datatest_url_files[1])
    if not Path(download_zip.as_posix()[:-4]).exists():
        if not download_zip.exists():
            click.secho(f"    Download data test\n", fg="yellow")
            multiprocessing_download([(snake_installer.datatest_url_files[0], download_zip.as_posix())], threads=1)
        click.secho(f"    Extract archive {download_zip} to {data_dir.as_posix()}\n", fg="yellow")
        unpack_archive(download_zip.as_posix(), data_dir.as_posix())
        download_zip.unlink()

    # build command line
    click.secho(f"    Write command line to run workflow on data test:\n", fg="yellow")
    mode = snake_installer.install_mode
    cmd = f"\n    {snake_installer.soft_name} {'run' if mode == 'cluster' else 'run --local-cores 1'} --configfile {data_config_path}\n\n"
    click.secho(cmd, fg='bright_blue')



from pathlib import Path

DOCS = "https://PKGNAME.readthedocs.io/en/latest/"
GIT_URL = "https://github.com/SouthGreenPlatform/PKGNAME"

SINGULARITY_URL_FILES = [('oras://registry.forge.ird.fr/diade/culebront_pipeline/apptainer/apptainer.culebront_tools.sif:0.0.1',
                          f'INSTALL_PATH/containers/apptainer.culebront_tools.sif')
                         ]

DATATEST_URL_FILES = ("https://itrop.ird.fr/culebront_utilities/Data-Xoo-sub.zip", "Data-Xoo-sub.zip")



